#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "sockev.h"
#include "sockev_internal.h"
#include "filters.h"

/* #define SOCKEV_DEBUG */

static P_inline void filter_call_out(sockev_t id, const char *data, int len);
static P_inline void filter_call_in(sockev_t id, char *data, int len);
static void P_inline filter_call_close(sockev_t id);

sockev_ctx_t sockev_ctx = {
	NULL,
	NULL,
	0,
	0,
	64,
	0
};

sockev_t sockev_alloc(void)
{
	int n;
	for(n = 0; n < USED; n++) {
		if (SOCKS[n].state == ST_FREE)
			goto found;
	}
	if (USED >= ALLOCED) {
		ALLOCED += GROW;
		PFDS    = realloc(PFDS,    sizeof(struct P_pollfd) * ALLOCED * 2);
		SOCKS = realloc(SOCKS, sizeof(sock_t) * ALLOCED);
	}
	n = USED;
	USED++;

	found:;
	SOCKS[n].state = ST_INUSE;
	SOCKS[n].s = P_NET_INVALID_SOCKET;
	SOCKS[n].sw = P_NET_INVALID_SOCKET;
	SOCKS[n].flags = 0;
	SOCKS[n].alive = 0;
	SOCKS[n].watch = 0;
	SOCKS[n].remote_port = 0;
	SOCKS[n].remote_ip = 0;
	SOCKS[n].remote_host = NULL;
	SOCKS[n].userdata = NULL;
	SOCKS[n].filters = 0;
	SOCKS[n].filt_cb = NULL;
	SOCKS[n].filt_data = NULL;
	SOCKS[n].closing = 0;
	SOCKS[n].final_in = NULL;
	SOCKS[n].final_out = NULL;
	SOCKS[n].final_data = NULL;
	SOCKS[n].final_shutdown = NULL;
	sockev_buffer_init(&SOCKS[n].wbuff);

	PFDS[n*2].fd = -1;
	PFDS[n*2].events = 0;
	PFDS[n*2].revents = 0;
	PFDS[n*2+1].fd = -1;
	PFDS[n*2+1].events = 0;
	PFDS[n*2+1].revents = 0;
	memset(&SOCKS[n].hooks, 0, sizeof(sockev_hooks_t));
	return n;
}

void sockev_free(sockev_t id)
{
	sock_t *s;
	s = &SOCKS[id];
	s->state = ST_FREE;
	s->s = P_NET_INVALID_SOCKET;
	s->sw = P_NET_INVALID_SOCKET;
	s->alive = 0;
	s->watch = 0;

	if (s->remote_host != NULL)
		free(s->remote_host);
	sockev_buffer_uninit(&s->wbuff);

	sockev_del_filters(id);

	PFDS[id*2].fd = -1;
	PFDS[id*2].events = 0;
	PFDS[id*2+1].fd = -1;
	PFDS[id*2+1].events = 0;
}

void sockev_free_all(void)
{
	int n;
	for(n = 0; n < USED; n++)
		if (SOCKS[n].state == ST_DEL)
			sockev_free(n);
}

static void sockev_update_fd(sockev_t id)
{
	if (PFDS[id*2].events == 0) {
		if (PFDS[id*2].fd > 0) {
			PFDS[id*2].fd = -PFDS[id*2].fd;
			PFDS[id*2].revents = 0;
		}
	}
	else {
		if (PFDS[id*2].fd < 0)
			PFDS[id*2].fd = -PFDS[id*2].fd;
	}

	if (PFDS[id*2+1].events == 0) {
		if (PFDS[id*2+1].fd > 0) {
			PFDS[id*2+1].fd = -PFDS[id*2+1].fd;
			PFDS[id*2+1].revents = 0;
		}
	}
	else {
		if (PFDS[id*2+1].fd < 0)
			PFDS[id*2+1].fd = -PFDS[id*2+1].fd;
	}
}

void sockev_del(sockev_t id)
{
	SOCKS[id].state = ST_DEL;
	PFDS[id*2].events = 0;
	PFDS[id*2+1].events = 0;
	sockev_update_fd(id);
}

static P_inline void sockev_error(sockev_t id, int shutdown)
{
	if (SOCKS[id].hooks.error != NULL)
		SOCKS[id].hooks.error(id);

	sockev_shutdown(id, shutdown);
}

/* check poll error on a specific pollfd (id is pollfd index) */
#define pollerr_(id) ((PFDS[id].revents & P_POLLHUP) || (PFDS[id].revents & P_POLLERR) || (PFDS[id].revents & P_POLLNVAL))

/* check poll error on a socket (id is socket index and both read and write sockets are checked if they differ) */
#define pollerr(id) (pollerr_(id*2) || ((SOCKS[id].sw != P_NET_INVALID_SOCKET) && pollerr_(id*2+1)))

static void handle_listen(sockev_t id)
{
	sock_t *s = &(SOCKS[id]);
	if (PFDS[id*2].revents & P_POLLIN) {
		if (s->hooks.readable != NULL)
			s->hooks.readable(id);
		else
			sockev_error(id, P_SHUT_RD);
	}
	if (pollerr(id))
		sockev_error(id, P_SHUT_RD);
}


static void handle_connect(sockev_t id)
{
	sock_t *s = &(SOCKS[id]);
	int pfd = id*2;

	if (s->sw != P_NET_INVALID_SOCKET)
		pfd++;

	if (pollerr_(pfd))
		sockev_error(id, P_SHUT_RDWR);
	else if (PFDS[pfd].revents & P_POLLOUT) {
		if (s->hooks.open != NULL)
			s->hooks.open(id);
		if (s->hooks.writable != NULL)
			s->hooks.writable(id);
		s->alive = SA_RD | SA_WR;
		s->watch = SA_RD | SA_WR;
		sockev_setpoll(id);
		s->state = ST_COMM;
	}
}

static P_inline int try_write(sockev_t id)
{
	const char *b;
	int len, wrote;

	if ((SOCKS[id].alive & SA_WR) == 0)
		return 0;

	if (sockev_buffer_get(&SOCKS[id].wbuff, &b, &len) == 0) {
		wrote = P_net_write((SOCKS[id].sw == P_NET_INVALID_SOCKET ? SOCKS[id].s : SOCKS[id].sw), (char *)b, len);
/*		printf("sockev WRITE: [%d] %d: %d bytes of %d bytes\n", id, SOCKS[id].sw == P_NET_INVALID_SOCKET ? SOCKS[id].s : SOCKS[id].sw, wrote, len);*/
		if (wrote < 0) {
			if (!P_net_inprogress) {
				/* write error */
				sockev_shutdown(id, P_SHUT_WR);
			}
		}
		else if (wrote > 0) {
			sockev_buffer_remove(&SOCKS[id].wbuff, wrote);
			return 1;
		}
	}
	return 0;
}

static void handle_comm(sockev_t id)
{
	sock_t *s = &(SOCKS[id]);
	int wpfd;

	wpfd = id*2;
	if (s->sw != P_NET_INVALID_SOCKET)
		wpfd++;

	/* handle pending write */
	if (PFDS[wpfd].revents & P_POLLOUT) {
		int noflush;
		noflush = 0;
		if (s->hooks.writable != NULL)
			noflush = s->hooks.writable(id);
		if (!noflush) {
			while(try_write(id)) ;
		}
		sockev_setpoll(id);
	}

	/* handle pending read */
	if (PFDS[id*2].revents & P_POLLIN) {
		int noread;
		noread = 0;
		if (s->hooks.readable != NULL)
			noread = s->hooks.readable(id);
		if (!noread) {
			char buff[1024];
			int len;
			len = P_net_read(s->s, buff, sizeof(buff)-1);
/*fprintf(stderr, "Read len: [%d] %d %d\n", id, len, errno);*/
			if (len == 0) {
				/* eof on read */
				sockev_shutdown(id, P_SHUT_RD);
			}
			else if (len < 0) {
				/* error on read */
				if (!P_net_inprogress) {
					sockev_error(id, P_SHUT_RD);
				}
#warning TODO
			}
			else {
				filter_call_in(id, buff, len);
			}
		}
	}
	if (pollerr(id))
		sockev_error(id, P_SHUT_WR);
}

int sockev_poll(int timeout_ms)
{
	int cnt;
	sockev_t n_;

	sockev_ctx.in_poll++;

	/* free all sockets scheduled for deletion before the call */
	sockev_free_all();
#ifdef SOCKEV_DEBUG
	fprintf(stderr, "Poll (%d): ", timeout_ms);
	for(n_ = 0; n_ < USED*2; n_++) {
		fprintf(stderr, "[%d]:%d:%d ", n_/2, PFDS[n_].fd, PFDS[n_].events);
	}
	fprintf(stderr, "\n");
#endif

	if (P_poll(PFDS, USED*2, timeout_ms) <= 0) {
#ifdef SOCKEV_DEBUG
	fprintf(stderr, "no poll\n");
#endif
		return 0;
	}

	cnt = 0;
	for(n_ = 0; n_ < USED*2; n_++) {
#warning TODO: HUP!
		sockev_t n;
		n = n_ / 2;
		if (PFDS[n_].fd < 0) {
			if (PFDS[n_].revents != 0)
				fprintf(stderr, "sockev error: poll event %d on unused SOCKET %d state %d (events=%d revents=%d fd=%d)\n", PFDS[n_].revents, n, SOCKS[n].state, PFDS[n_].events, PFDS[n_].revents, PFDS[n_].fd);
			continue;
		}
		if (PFDS[n_].revents == 0) {
			continue;
		}
#ifdef SOCKEV_DEBUG
		fprintf(stderr, " Event %d on %d state=%d\n", PFDS[n_].revents, n, SOCKS[n].state);
#endif
		switch(SOCKS[n].state) {
			case ST_LISTEN:  handle_listen(n); cnt++; break;
			case ST_CONNECT: handle_connect(n); cnt++; break;
			case ST_COMM:    handle_comm(n); cnt++; break;
			case ST_DEL: /* deleted while in the loop - ignore */ break;
			case ST_VIRTUAL:
				fprintf(stderr, "sockev error: poll event %d on virtual socket %d state %d\n", PFDS[n_].revents, n, SOCKS[n].state);
				PFDS[n_].events = 0; /* disable events to avoid further triggers and short polls */
				sockev_update_fd(n_);
				break;
			case ST_INUSE:
			case ST_FREE:
				/* event on a socket not in use yet - this should not happen */
				fprintf(stderr, "sockev error: poll event %d on unused socket %d state %d (events=%d)\n", PFDS[n_].revents, n, SOCKS[n].state, PFDS[n_].events);
				PFDS[n_].events = 0; /* disable events to avoid further triggers and short polls */
				sockev_update_fd(n_);
				break;
		}
	}

	sockev_ctx.in_poll--;


	/* free all sockets scheduled for deletion */
	if (sockev_ctx.in_poll == 0)
		sockev_free_all();
	return cnt;
}

void sockev_write_(sockev_t id, const char *buff, int len)
{
	filter_call_out(id, buff, len);
}

void sockev_write(sockev_t id, const char *buff, int len)
{
	if (wrong_id(id, -1))
		return;
	if ((SOCKS[id].alive & SA_WR) == 0)
		return;
	sockev_write_(id, buff, len);
}

void sockev_puts(sockev_t id, const char *str)
{
	if (wrong_id(id, -1))
		return;
	if ((SOCKS[id].alive & SA_WR) == 0)
		return;
	sockev_write_(id, str, strlen(str));
}

int sockev_read(sockev_t id, char *buff, int len)
{
	if (wrong_id(id, ST_COMM))
		return -2;
	return P_net_read(SOCKS[id].s, buff, len);
}

static void sockev_close_(sockev_t id)
{
	if (SOCKS[id].state == ST_DEL)
		return;
	filter_call_close(id);
	if ((SOCKS[id].hooks.close != NULL) && (!SOCKS[id].closing)) {
		SOCKS[id].closing = 1;
		SOCKS[id].hooks.close(id);
	}
#ifdef SOCKEV_DEBUG
	fprintf(stderr, "sockev: close %d: %d %d\n", id, SOCKS[id].s, SOCKS[id].sw);
#endif
	if (SOCKS[id].s >= 0)
		P_net_close(SOCKS[id].s);
	if (SOCKS[id].sw >= 0)
		P_net_close(SOCKS[id].sw);
	sockev_del(id);
}

void sockev_shutdown(sockev_t id, int direction)
{
	int new_ev;

	if (wrong_id(id, -1))
		return;

	if ((SOCKS[id].state != ST_CONNECT) && (SOCKS[id].state != ST_COMM) && (SOCKS[id].state != ST_LISTEN)) {
#ifdef SOCKEV_DEBUG
		fprintf(stderr, "sockev: shutdown fail %d: state is %d\n", id, SOCKS[id].state);
#endif
		return;
	}

	/* if we don't have half-close for this socket, close it fully */
	if (SOCKS[id].flags & SEVFLG_NO_HALF_CLOSE)
		direction = P_SHUT_RDWR;

	if (SOCKS[id].final_shutdown != NULL) {
		SOCKS[id].final_shutdown(id, direction);
	}
	else {
		P_net_shutdown(SOCKS[id].s, direction);
		if (SOCKS[id].sw != P_NET_INVALID_SOCKET)
			P_net_shutdown(SOCKS[id].sw, direction);
	}

	new_ev = 0;
	if ((direction == P_SHUT_WR) || (direction == P_SHUT_RDWR)) {
		/* note if this is a new close event */
		if (SOCKS[id].alive & SA_WR) {
			SOCKS[id].alive &= ~SA_WR;
			new_ev++;
		}
	}

	if ((direction == P_SHUT_RD) || (direction == P_SHUT_RDWR)) {
			/* note if this is a new close event */
			if (SOCKS[id].alive & SA_RD) {
				SOCKS[id].alive &= ~SA_RD;
				new_ev++;
/*				fprintf(stderr, "Set poll %d %d cb=%p!\n", id, PFDS[id*2].events, SOCKS[id].hooks.readeof);*/
			}
			if ((new_ev) && (SOCKS[id].hooks.readeof != NULL))
				SOCKS[id].hooks.readeof(id);
	}

	sockev_setpoll(id);
	if (SOCKS[id].alive == 0)
		sockev_close_(id);
}

void sockev_init(void)
{
	P_net_init();
}

void sockev_uninit(void)
{
	int n;
	for(n = 0; n < USED; n++)
		sockev_close(n);
	sockev_free_all();
	free(PFDS);
	free(SOCKS);
	USED = ALLOCED = 0;
	P_net_uninit();
}


void sockev_set_userdata(sockev_t id, void *ud)
{
	if (wrong_id(id, -1))
		return;
	SOCKS[id].userdata = ud;
}

void *sockev_get_userdata(sockev_t id)
{
	if (wrong_id(id, -1))
		return NULL;
	return SOCKS[id].userdata;
}


sockev_t sockev_cfg_hooks(sockev_t id, const sockev_hooks_t *hooks)
{
	if (id == SOCKEV_NEW)
		id = sockev_alloc();
	else if (wrong_id(id, -1))
		return -1;

	memcpy(&SOCKS[id].hooks, hooks, sizeof(sockev_hooks_t));
	return id;
}

sockev_t sockev_cfg_flags(sockev_t id, sockev_flags_t flags)
{
	if (id == SOCKEV_NEW)
		id = sockev_alloc();
	else if (wrong_id(id, -1))
		return -1;

	SOCKS[id].flags = flags;
	return id;
}

void sockev_setfds(sockev_t id, P_net_socket fdi, P_net_socket fdo)
{
	SOCKS[id].s = fdi;
	SOCKS[id].sw = fdo;
	PFDS[id*2].fd = fdi;
	PFDS[id*2+1].fd = fdo;
}

void sockev_setpoll(sockev_t id)
{
	PFDS[id*2].events = 0;
	PFDS[id*2+1].events = 0;

	if (SOCKS[id].state == ST_VIRTUAL) {
		sockev_update_fd(id);
		return; /* never poll on virtual sockets */
	}

	/* poll for out, if enabled, not closed and have data or waiting for connect */
	if ((SOCKS[id].watch & SA_WR) && (SOCKS[id].alive & SA_WR) && ((sockev_buffer_total(&SOCKS[id].wbuff) != 0) || (SOCKS[id].state == ST_CONNECT))) {
		if (SOCKS[id].sw != P_NET_INVALID_SOCKET)
			PFDS[id*2+1].events |= P_POLLOUT;
		else
			PFDS[id*2].events |= P_POLLOUT;
	}

	/* poll for in if enabled and not closed */
	if ((SOCKS[id].watch & SA_RD) && (SOCKS[id].alive & SA_RD))
		PFDS[id*2].events |= P_POLLIN;
#ifdef SOCKEV_DEBUG
	fprintf(stderr, "setpoll %d %d/%d\n", id, PFDS[id*2].events, PFDS[id*2+1].events);
#endif

	sockev_update_fd(id);
}


sockev_t sockev_set_watch(sockev_t id, int read, int write)
{
	if (wrong_id(id, -1))
		return SOCKEV_INVALID;

	SOCKS[id].watch = (read ? SA_RD : 0) | (write ? SA_WR : 0);
	sockev_setpoll(id);
	return id;
}

int sockev_get_remote(sockev_t id, P_ipv4_addr_t *ip, const char **hostname, int *port)
{
	if (wrong_id(id, -1))
		return -1;

	if (ip != NULL)
		*ip = SOCKS[id].remote_ip;

	if (hostname != NULL)
		*hostname = SOCKS[id].remote_host;

	if (port != NULL)
		*port = SOCKS[id].remote_port;

	return 0;
}

int sockev_get_wbuf(sockev_t id)
{
	if (wrong_id(id, -1))
		return -1;

	return sockev_buffer_total(&SOCKS[id].wbuff);
}

static void P_inline filter_call_out(sockev_t id, const char *data, int len)
{
	if (SOCKS[id].filters == 0)
		sockev_filter_final_out(id, NULL, SOCKEV_F_OUT, data, len, 0);
	else
		sockev_filter_next(id, SOCKEV_F_OUT, (void *)data, len, -1); /* we trust filters not to modify data in OUT */
}

static void P_inline filter_call_in(sockev_t id, char *data, int len)
{
	if (SOCKS[id].filters == 0)
		sockev_filter_final_in(id, NULL, SOCKEV_F_IN, data, len, 0);
	else
		sockev_filter_next(id, SOCKEV_F_IN, data, len, SOCKS[id].filters);
}

static void P_inline filter_call_close(sockev_t id)
{
	if (SOCKS[id].filters != 0)
		sockev_filter_next(id, SOCKEV_F_CLOSE, NULL, 0, SOCKS[id].filters);
}

/* published symbols for special filters */
void sockev_filter_call_out(sockev_t id, const char *data, int len)
{
	filter_call_out(id, data, len);
}

void sockev_filter_call_in(sockev_t id, char *data, int len)
{
	filter_call_in(id, data, len);
}

void sockev_filter_call_close(sockev_t id)
{
	filter_call_close(id);
}


/* we can't use fdopen as an alternative because fclose() would close the
socket. We could hash P_net_socket -> FILE *, but we wouldn't notice
if the fd was replaced meanwhile and the system will most probably
reuse fd numbers. (Or else all libporty calls that could affect what an fd
means would need to maintain the hash). */
#include <stdlib.h>
#include <stdarg.h>
#include "libporty/config.h"

#ifdef P_HAVE_VSNPRINTF
#	ifdef P_HAVE_VSNPRINTF_SAFE
#	define SAFETY 0
#	else
#	define SAFETY 1
#	endif
void sockev_printf(sockev_t id, const char *format, ...)
{
	int size;
	char *buff;
	int ret;
	va_list ap;

	if (wrong_id(id, -1))
		return;

	for(size = P_net_printf_minbuff; size <= P_net_printf_maxbuff; size = size * 2) {
		buff = malloc(size + SAFETY);
		if (buff == NULL)
			return 0;
		va_start(ap, format);
		ret = vsnprintf(buff, size, format, ap);
		va_end(ap);
		if (ret <= size && ret != -1)
			break;
	}
	if (ret <= size && ret >= 0)
		sockev_write(id, buff, ret);

	free(buff);
}
#else
/* we don't have anything advanced so allocate a 'large enough' buffer and
check for overflow */
#include <stdlib.h>
#include <stdarg.h>
void sockev_printf(sockev_t id, const char *format, ...)
{
	char *buff;
	int len;
	va_list ap;
	va_start(ap, format);

	if (wrong_id(id, -1))
		return;

	buff = malloc(P_net_printf_maxbuff);
	len = vsprintf(buff, format, ap);
	if (len >= P_net_printf_maxbuff) {
		fprintf(stderr, "sockev_printf: buffer (%d) too small (need %d); please increase P_net_printf_maxbuff\n", P_net_printf_maxbuff, len);
		abort();
	}
	sockev_write(id, buff, ret);

	free(buff);
	va_end(ap);
	return len;
}
#endif

